package meerkat.bulletinboard;

import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.ListeningScheduledExecutorService;
import com.google.common.util.concurrent.MoreExecutors;
import com.google.protobuf.ByteString;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.Executors;

/**
 * Created by Arbel Deutsch Peled on 03-Mar-16.
 * This is a full-fledged implementation of a Bulletin Board Client
 * It provides asynchronous access to several remote servers, as well as a local cache
 * Read/write operations are performed on the local server
 * After any read is carried out, a subscription is made for the specific query to make sure the local DB will be updated
 * The database also employs a synchronizer which makes sure local data is sent to the remote servers
 */
public class CachedBulletinBoardClient implements SubscriptionAsyncBulletinBoardClient {

    private final BulletinBoardClient localClient;
    private AsyncBulletinBoardClient remoteClient;
    private BulletinBoardSubscriber subscriber;

    private final int threadPoolSize;
    private final long failDelayInMilliseconds;
    private final long subscriptionIntervalInMilliseconds;

    public CachedBulletinBoardClient(BulletinBoardClient localClient,
                                     int threadPoolSize,
                                     long failDelayInMilliseconds,
                                     long subscriptionIntervalInMilliseconds)
            throws IllegalAccessException, InstantiationException {

        this.localClient = localClient;
        this.threadPoolSize = threadPoolSize;
        this.failDelayInMilliseconds = failDelayInMilliseconds;
        this.subscriptionIntervalInMilliseconds = subscriptionIntervalInMilliseconds;

        remoteClient = new ThreadedBulletinBoardClient();

    }

    @Override
    public MessageID postMessage(BulletinBoardMessage msg, FutureCallback<Boolean> callback) {
        return null;
    }

    @Override
    public MessageID postBatch(CompleteBatch completeBatch, FutureCallback<Boolean> callback) {
        return null;
    }

    @Override
    public void beginBatch(BeginBatchMessage beginBatchMessage, FutureCallback<Boolean> callback) {

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList, int startPosition, FutureCallback<Boolean> callback) {

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList, FutureCallback<Boolean> callback) {

    }

    @Override
    public void postBatchData(ByteString signerId, int batchId, List<BatchData> batchDataList, int startPosition, FutureCallback<Boolean> callback) {

    }

    @Override
    public void postBatchData(ByteString signerId, int batchId, List<BatchData> batchDataList, FutureCallback<Boolean> callback) {

    }

    @Override
    public void closeBatch(CloseBatchMessage closeBatchMessage, FutureCallback<Boolean> callback) {

    }

    @Override
    public void getRedundancy(MessageID id, FutureCallback<Float> callback) {

    }

    @Override
    public void readMessages(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {

    }

    @Override
    public void readBatch(BatchSpecificationMessage batchSpecificationMessage, FutureCallback<CompleteBatch> callback) {

    }

    @Override
    public void querySync(SyncQuery syncQuery, FutureCallback<SyncQueryResponse> callback) {

    }

    @Override
    public void init(BulletinBoardClientParams clientParams) {

        remoteClient.init(clientParams);

        ListeningScheduledExecutorService executorService = MoreExecutors.listeningDecorator(Executors.newScheduledThreadPool(threadPoolSize));

        List<SubscriptionAsyncBulletinBoardClient> subscriberClients = new ArrayList<>(clientParams.getBulletinBoardAddressCount());

        for (String address : clientParams.getBulletinBoardAddressList()){

            SubscriptionAsyncBulletinBoardClient newClient =
                    new SingleServerBulletinBoardClient(executorService, failDelayInMilliseconds, subscriptionIntervalInMilliseconds);

            newClient.init(clientParams.toBuilder().clearBulletinBoardAddress().addBulletinBoardAddress(address).build());

            subscriberClients.add(newClient);

        }

        subscriber = new ThreadedBulletinBoardSubscriber(subscriberClients, localClient);

    }

    @Override
    public MessageID postMessage(BulletinBoardMessage msg) throws CommunicationException {
        return null;
    }

    @Override
    public float getRedundancy(MessageID id) {
        return 0;
    }

    @Override
    public List<BulletinBoardMessage> readMessages(MessageFilterList filterList) {
        return null;
    }

    @Override
    public SyncQuery generateSyncQuery(GenerateSyncQueryParams GenerateSyncQueryParams) throws CommunicationException {
        return null;
    }

    @Override
    public void close() {

    }

    @Override
    public void subscribe(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {

    }

    @Override
    public void subscribe(MessageFilterList filterList, long startEntry, FutureCallback<List<BulletinBoardMessage>> callback) {

    }
}
