package meerkat.crypto.dkg.comm;

import com.google.protobuf.Message;
import meerkat.crypto.utils.Channel;
import meerkat.protobuf.DKGMessages;

/**
 * Created by Tzlil on 2/14/2016.
 *
 * an implementation of ReceiverCallback
 */
public abstract class MailHandler implements Channel.ReceiverCallback{

    /**
     * fixed value for broadcasting
     */
    public static final int BROADCAST = 0;

    /**
     * message handler
     */
    private MessageHandler messageHandler;

    /**
     * constructor
     * @param messageHandler
     */
    public MailHandler(MessageHandler messageHandler){
        this.messageHandler = messageHandler;
    }

    /**
     * extract message from mail
     * @param mail
     * @return
     */
    public abstract Message extractMessage(DKGMessages.Mail mail);

    /**
     * is this mail was received by broadcast channel
     * @param mail
     * @return mail user destination == BROADCAST
     */
    public boolean isBroadcast(DKGMessages.Mail mail){
        return mail.getDestination() == BROADCAST;
    }

    @Override
    public void receiveMail(DKGMessages.Mail mail){
        Message message = extractMessage(mail);
        if (message == null)
            return;

        switch (mail.getType()) {
            case SHARE:
                messageHandler.handleShareMessage(mail.getSender(), isBroadcast(mail),message);
                break;
            case COMMITMENT:
                messageHandler.handleCommitmentMessage(mail.getSender(), isBroadcast(mail),message);
                break;
            case DONE:
                messageHandler.handleDoneMessage(mail.getSender(), isBroadcast(mail),message);
                break;
            case COMPLAINT:
                messageHandler.handleComplaintMessage(mail.getSender(), isBroadcast(mail),message);
                break;
            case ANSWER:
                messageHandler.handleAnswerMessage(mail.getSender(), isBroadcast(mail),message);
                break;
            case ABORT:
                messageHandler.handleAbortMessage(mail.getSender(), isBroadcast(mail),message);
                break;
            default:
                break;
        }


    }
}
