package meerkat.crypto.utils;

import com.google.protobuf.Message;
import meerkat.protobuf.DKGMessages;

/**
 * A generic communication channel that supports point-to-point and broadcast operation
 */

public interface Channel {
    public interface ReceiverCallback {
        public void receiveMail(DKGMessages.Mail mail);
    }

    /**
     * sends a private message
     * @param destUser destination user's identifier
     * @param type message type
     * @param msg message
     */
    public void sendMessage(int destUser, DKGMessages.Mail.Type type, Message msg);


    /**
     * broadcasts a message to all parties (including the sender)
     * @param type message type
     * @param msg message
     */
    public void broadcastMessage(DKGMessages.Mail.Type type, Message msg);

    /**
     * Register a callback to handle received messages.
     * The callback is called in the <b>Channel</b> thread, so no long processing should
     * occur in the callback method.
     * @param callback
     */
    public void registerReceiverCallback(ReceiverCallback callback);
}
